print("import: 'chainer'")
import chainer

print("import: 'chainer.backends'")
import chainer.backends

print("import: 'chainer.dataset'")
import chainer.dataset

print("import: 'chainer.datasets'")
import chainer.datasets

print("import: 'chainer.distributions'")
import chainer.distributions

print("import: 'chainer.exporters'")
import chainer.exporters

print("import: 'chainer.functions'")
import chainer.functions

print("import: 'chainer.functions.activation'")
import chainer.functions.activation

print("import: 'chainer.functions.array'")
import chainer.functions.array

print("import: 'chainer.functions.connection'")
import chainer.functions.connection

print("import: 'chainer.functions.evaluation'")
import chainer.functions.evaluation

print("import: 'chainer.functions.loss'")
import chainer.functions.loss

print("import: 'chainer.functions.math'")
import chainer.functions.math

print("import: 'chainer.functions.noise'")
import chainer.functions.noise

print("import: 'chainer.functions.normalization'")
import chainer.functions.normalization

print("import: 'chainer.functions.pooling'")
import chainer.functions.pooling

print("import: 'chainer.functions.theano'")
import chainer.functions.theano

print("import: 'chainer.functions.util'")
import chainer.functions.util

print("import: 'chainer.function_hooks'")
import chainer.function_hooks

print("import: 'chainer.iterators'")
import chainer.iterators

print("import: 'chainer.initializers'")
import chainer.initializers

print("import: 'chainer.graph_optimizations'")
import chainer.graph_optimizations

print("import: 'chainer.optimizers'")
import chainer.optimizers

print("import: 'chainer.optimizer_hooks'")
import chainer.optimizer_hooks

print("import: 'chainer.serializers'")
import chainer.serializers

print("import: 'chainer.training'")
import chainer.training

print("import: 'chainer.training.extensions'")
import chainer.training.extensions

print("import: 'chainer.training.triggers'")
import chainer.training.triggers

print("import: 'chainer.training.updaters'")
import chainer.training.updaters

print("import: 'chainer.utils'")
import chainer.utils

print("import: 'chainermn'")
import chainermn

print("import: 'chainermn.communicators'")
import chainermn.communicators

print("import: 'chainermn.datasets'")
import chainermn.datasets

print("import: 'chainermn.extensions'")
import chainermn.extensions

print("import: 'chainermn.functions'")
import chainermn.functions

print("import: 'chainermn.iterators'")
import chainermn.iterators

